/*
 * Copyright (c) 2009 The Trustees of Indiana University and Indiana
 *                    University Research and Technology
 *                    Corporation.  All rights reserved.
 *
 * Author(s): Torsten Hoefler <htor@cs.indiana.edu>
 *
 */

#ifndef __NBCBENCH_H__
#define __NBCBENCH_H__

/* define synchronization method here */
//#define SYNC_WINDOW
#define SYNC_BARRIER

/* undefine for MPI only benchmark */
#define HAVE_NBC

/* number of warmup operations for every datasize/comm-size combination */
#define NBCB_WARMUP 1

/* compute time factor - how much longer than the NBC latency should be computed? */
#define NBC_COMPTIMEFACT 2.0

#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

#ifdef HAVE_NBC
#include <nbc.h>
#endif

/* result array elements */
#define NBC 0
#define NBCCOMPOV 1
#define NBCCOMPTEST 2
#define NBCINIT 3
#define NBCTEST 4
#define RES_MPI 5
#define NBCCOMPWAIT 6
/* this is the number of benches in the inner loop - basically the
 * number of elements in the time-array 
 * It should only be changed if new benches are added */
#define NUM_BENCH 7

/* the operations */
#define ALLGATHER 1
#define ALLGATHERV 2
#define ALLREDUCE 3
#define ALLTOALL 4
#define ALLTOALLV 5
#define ALLTOALLW 6
#define BARRIER 7
#define BCAST 8
#define EXSCAN 9
#define GATHER 10
#define GATHERV 11
#define REDUCE 12
#define REDUCESCAT 13
#define SCAN 14
#define SCATTER 15
#define SCATTERV 16
#define SEND 17
#define NEIGHBOR_XCHG 18
#define NEIGHBOR_XCHGV 19

typedef enum {MIN, MAX, MED, AVG} t_funcs;
typedef enum {COMP_TIME, COMP_WORK} t_compfunc;

typedef struct {
  int rank; /* MPI rank */
  char verbose; /* -v ? */
  char printall; /* -a ? */
  char op; /* -t */
  int pmin; /* -p */
  int pmax; /* -p */
  int smin; /* -s */
  int smax; /* -s */
  int trials; /* -n */
  int testint; /* -i */
  int numneighbor; /* -x */
  FILE *output; /* -o (FILE for output) */
  int do_output; /* do output? 0/1 */
  t_funcs red; /* -r */
  t_compfunc compfunc; /* -c */
} t_flags;  

typedef struct {
  void *buf;
  int size;
  int sizered;
  void *rbuf;
  int *rcounts;
  int *rcountsred; /* rcounts for reduce operations */
  int *scounts;
  int *rdispls;
  int *sdispls;
  int *indexes; /* for sparse communication - see graph definition */
  int *edges; /* for sparse communication - see graph definition */
  int rank;
  int nprocs;
  MPI_Comm comm;
  MPI_Comm graphcomm;

  t_compfunc compfunc; /* -c */
  double computetime; /* time needed for computing those loops in usec */
  unsigned long long computecycles; /* number of computecycles */
  unsigned long tests; /* number of tests per computation */
  
#ifdef HAVE_NBC
  NBC_Handle *handle;
#endif
  double esttime; /* estimated maximum single step time (=max(estnbctime, estmpitime)) in usec */
#ifdef SYNC_WINDOW
  double window; /* window to perform operation */
#endif
} t_mpiargs;

static __inline__ double mea_mpi(t_flags *flags, t_mpiargs *mpiargs);
static __inline__ double mea_nbccomp(t_flags *flags, t_mpiargs *mpiargs);
static __inline__ double mea_nbc(t_flags *flags, t_mpiargs *mpiargs);
static __inline__ double mea_nbccompov(t_flags *flags, t_mpiargs *mpiargs, double *testtime, double *waittime);
static __inline__ double mea_nbcinit(t_flags *flags, t_mpiargs *mpiargs, double *testtime);
static int do_benchmark(t_flags flags);
void option_usage(char *opt, char *desc, char *param);

/* synchronization functions */
void sync_init_stage1(t_mpiargs *mpiargs);
void sync_init_stage2(t_mpiargs *mpiargs);
double nbcb_sync(t_mpiargs *mpiargs);


/* ok - uint64_t needs GCC??? */
#ifdef __PGI
typedef unsigned long long u_int64_t;
#endif

/* restore inline behavior for non-gcc compilers */
#ifndef __GNUC__
#define __inline__ inline
#endif

#define MAX_DOUBLE 1e99

#endif
